import logging
import transaction
from ringo.lib.helpers import dynamic_import
from ringo.lib.imexport import Importer
from ringo.model.modul import ModulItem
from euhu.programs.ec import validate
from ecimporter.lib.data_validation import check_participation_form
from ecimporter.lib.data_validation import check_participant_form


log = logging.getLogger(__name__)


class FormValidationError(Exception):
    """Show case, if a form is not validated."""
    pass


class DictImporter(Importer):
    """Import from a python dictionary with serialized values."""

    def deserialize(self, data):
        result = []
        for d in data:
            result.append(self._deserialize_values(d))
        return result


def get_esf_validation(pt_dict, pn_dict, oi_dict):
    """Validate ESF rules."""
    # build the data
    data_dict = {}
    data_dict.update(pt_dict)
    data_dict.update(pn_dict)
    data_dict.update(oi_dict)

    # validate with euhu function
    validation_result = validate(data_dict)
    if validation_result.has_violated_rules():
        return 0
    else:
        return 1


def get_logging_from_participants(items):
    return {"participants": [(item[0].id, item[0].personal_id,
                              item[0].owner.login)
                             for item in items]}


def get_logging_from_participations(items):
    return {"participations": [item[0].id for item in items]}


def get_logging_from_request(request):
    return {"db": request.registry.settings["sqlalchemy.url"].split("/")[-1]}


def ecimport(data, request):
    session = request.db
    try:
        pts = [d["participant"] for d in data]
        pns = [d["participation"] for d in data]
        ois = [d["oi"] for d in data]

        for pt in pts:
            if not check_participant_form(pt, request):
                raise FormValidationError("Form participant doesn't validate")

        for n, _ in enumerate(pts):
            if not check_participation_form(pts[n], pns[n], ois[n], request):
                raise FormValidationError(
                    "Form participation doesn't validate")
        transaction.begin()
        items, pt_created, pt_updated = do_import(session, 'participants',
                                                  pts)
        session.flush()
        logging_info = get_logging_from_participants(items)

        set_ptid_in_pns(items, pns)
        pn_items, pn_c, pn_u = do_import(session, 'participations', pns)
        session.flush()
        logging_info.update(get_logging_from_participations(pn_items))
        logging_info.update(get_logging_from_request(request))

        set_id_in_ois(pn_items, ois)
        _, oi_c, oi_u = do_import(session, 'outputindicators', ois)
        for n, _ in enumerate(pts):
            pn_items[n][0].valid = get_esf_validation(pts[n], pns[n], ois[n])
            log.debug("Set participation valid to %d", pn_items[n][0].valid)

    except FormValidationError as exp:
        log.error("Form Validation failed")
        log.error(exp.message)
        raise exp
    except (AttributeError, Exception) as exp:
        transaction.abort()
        log.exception("Abort Transaction")
        raise exp
    else:
        transaction.commit()
        for participant_log_info, participation_id in \
                zip(logging_info["participants"],
                        logging_info["participations"]):
            log.info("%s: Imported Participant %s as %s. Participation"
                     " is %s (User: %s)",
                     logging_info["db"],
                     participant_log_info[1],
                     participant_log_info[0],
                     participation_id,
                     participant_log_info[2])
        log.debug("Participant: Updated %s items, Created %s items",
                 pt_updated, pt_created)
        log.debug("Participation: Updated %s items, Created %s items",
                 pn_u, pn_c)
        log.debug("OI: Updated %s items, Created %s items", oi_u, oi_c)


def get_importer(session, modulname):
    modul_clazzpath = session.query(ModulItem).filter(
        ModulItem.name == modulname).all()[0].clazzpath
    modul = dynamic_import(modul_clazzpath)
    return DictImporter(modul, session)


def set_ptid_in_pns(items, pns):
    for item, pn in zip(items, pns):
        pn["participant_id"] = item[0].id
        pn["project_id"] = 1


def set_id_in_ois(items, ois):
    for item, oi in zip(items, ois):
        oi["id"] = item[0].id


def do_import(session, importer_module, data):
    items = []
    updated = 0
    created = 0
    importer = get_importer(session, importer_module)
    items = importer.perform(data, load_key="id")
    for item, action in items:
        # Add all new items to the session
        if action.find("CREATE") > -1:
            session.add(item)
            created += 1
        else:
            updated += 1
    return items, created, updated
